/*
 * Number decoder
 *
 * Written by Toshiharu Ohno (sirius506@gmail.com)
 */
#include <stdint.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include "SAMCamera.h"
#include "print.h"

#define	USE_VECTOR

#define	IMG_WIDTH  160
#define IMG_HEIGHT 120

/* Our dictionary bitmap size is 8 X 16 pixels */
#define	NORM_WIDTH 8
#define	NORM_HEIGHT 16

#include "digitdef.h"

#define	BLACK_PIXEL	0x00
#define	WHITE_PIXEL	0xff

typedef union {
  struct s_pl {
    uint16_t pixel:1;
    uint16_t label:15;
  } __attribute__((packed)) pix_label;
  uint8_t rgb[2];
} __attribute__((packed)) IMAGEDATA;

uint8_t normed_digit[NORM_WIDTH*NORM_HEIGHT];
#ifdef USE_VECTOR
short digit_vector[NORM_WIDTH+NORM_HEIGHT];
#endif
uint8_t raster[IMG_WIDTH*2];
uint8_t label_cleared[IMG_WIDTH*IMG_HEIGHT/8/4];

#define	N_HIST	32

short histogram[N_HIST];
char  histmark[N_HIST];

typedef struct {
  short xpos, ypos;
  short width, height;
  short xbars, ybars;
  short area;
} LABELPROP;

void label_clear(IMAGEDATA *base, int number, LABELPROP *pProp)
{
  IMAGEDATA *img;
  int w, h;

  for (h = 0; h < pProp->height; h++) {
    img = base + IMG_WIDTH * (pProp->ypos + h) + pProp->xpos;
    for (w = 0; w < pProp->width; w++) {
      if (img->pix_label.label == number) {
        img->pix_label.label = 0;
        img->pix_label.pixel = 0;
      }
      img++;
    }
  }
}

void bmp_write(IMAGEDATA *image)
{
  int x, y;
  uint8_t pixel, *pRaster;

  for (y = 0; y < IMG_HEIGHT; y++) {
    pRaster = raster;
    for (x = 0; x < IMG_WIDTH; x++) {
      pixel = image->pix_label.pixel? BLACK_PIXEL : WHITE_PIXEL;
      *pRaster++ = pixel;
      *pRaster++ = pixel;
      image++;
    }
    lcd_copy_line(raster, y);
  }
}

int labeling(IMAGEDATA *pImg)
{
  int x, y, x2, y2;
  int lambda, lambda1, lambda2;
  int l1, l2, l3, l4;
  int sw;
  IMAGEDATA *wp;

  lambda = 0;

  for (y = 1; y < IMG_HEIGHT; y++) {
    for (x = 1; x < IMG_WIDTH - 1; x++) {
      wp = pImg + IMG_WIDTH * y + x;
      if (wp->pix_label.pixel) {	// Black
        l1 = wp[-IMG_WIDTH-1].pix_label.label;
        l2 = wp[-IMG_WIDTH].pix_label.label;
        l3 = wp[-IMG_WIDTH+1].pix_label.label;
        l4 = wp[-1].pix_label.label;

        if (l1 == 0 && l2 == 0 && l3 == 0 && l4 == 0) {
          lambda++;
          wp[0].pix_label.label = lambda;
        } else {
          lambda1 = l1;
          lambda2 = -1;
          if (l1 != l2) lambda2 = l2;
          else if (l1 != l3) lambda2 = l3;
          else if (l1 != l4) lambda2 = l4;
          if (lambda1 > lambda2) {
            sw = lambda1;
            lambda1 = lambda2;
            lambda2 = sw;
          }
          if (lambda1 == -1 || lambda1 == 0) {
            wp[0].pix_label.label = lambda2;
          } else {
            wp[0].pix_label.label = lambda1;
            for (y2 = 0; y2 <= y; y2++) {
              for (x2 = 0; x2 < IMG_WIDTH; x2++) {
                if (pImg[IMG_WIDTH*y2+x2].pix_label.label == lambda2)
                  pImg[y2*IMG_WIDTH+x2].pix_label.label = lambda1;
              }
            }
            /* lambda2 label is no longer used */
            label_cleared[lambda2/8] |= (1 << (lambda2&7));
          }
        }
      } else {
        wp[0].pix_label.label = 0;
      }
    }
  }
  return lambda;
}

void get_label_property(IMAGEDATA *pImg, int label, LABELPROP *pProp)
{
  int minx, miny, maxx, maxy;
  int x, y;

  minx = IMG_WIDTH;
  miny = IMG_HEIGHT;
  maxx = maxy = 0;
  pProp->xpos = pProp->ypos = 0;
  pProp->width = pProp->height = 0;
  pProp->area = 0;
  for (y = 0; y < IMG_HEIGHT;y++) {   /* Check every pixels in a region */
    for (x = 0; x < IMG_WIDTH; x++) {
      if (pImg->pix_label.label == label) {
        if (x < minx) minx = x;
        if (x > maxx) maxx = x;
        if (y < miny) miny = y;
        if (y > maxy) maxy = y;
        pProp->area++;
      }
      pImg++;
    }
  }
  if (pProp->area > 0) {
    pProp->xpos = minx;
    pProp->ypos = miny;
    pProp->width = maxx - minx + 1;
    pProp->height = maxy - miny + 1;
#ifdef LDEBUG
    cons_printf("%d: (%d, %d) --> %d X %d\r\n", label, pProp->xpos, pProp->ypos, pProp->width, pProp->height);
#endif
  }
}

typedef int FIX_FLOAT;
#define	FIX_HALF	0x80	/* 0.5 */

int interpolate(IMAGEDATA *pImg, FIX_FLOAT x, FIX_FLOAT y, int *pixval)
{
  int px, py;

  px = (x+FIX_HALF) >> 8;
  py = (y+FIX_HALF) >> 8;
  if (px < 0 || px >= IMG_WIDTH)
    return 0;
  if (py < 0 || py >= IMG_HEIGHT)
    return 0;
  pImg += (py * IMG_WIDTH) + px;
  *pixval = (pImg->pix_label.pixel)? 0xff : 0x00;
  return 1;
}

/*
 * Resize given region image into NORM_WIDTH by NORM_HEIGHT pixels.
 */
void resize_region(IMAGEDATA *pImg, LABELPROP *pProp)
{
  int x, y;
  FIX_FLOAT fx, fy;
  int pixval;
  uint8_t *normImg;

  memset(normed_digit, 0, sizeof(normed_digit));
  normImg = normed_digit;

  for (y = 0; y < NORM_HEIGHT; y++) { /* For each pixels in a digit */
    for (x = 0; x < NORM_WIDTH; x++) {
      /* Compute corresponding location on the original image */
      fx = ((x* pProp->width) << 8)/(NORM_WIDTH << 8)+ pProp->xpos;
      fx <<= 8;
      fy = ((y*pProp->height) << 8)/(NORM_HEIGHT << 8)+ pProp->ypos;
      fy <<= 8;
      /* pick a pixel value on the original image */
      if (interpolate(pImg, fx, fy, &pixval)) {
        *normImg = pixval;
      }
      normImg++;
    }
  }
}

#ifdef USE_VECTOR
void compute_vector()
{
  int x, y, black;
  uint8_t *src;

  src = normed_digit;
  /* Count black pixels in a column */
  for (y = 0; y < NORM_HEIGHT; y++) {
    black = 0;
    for (x = 0; x < NORM_WIDTH; x++) {
      if (*src) black++;
      src++;
    }
    digit_vector[y] = black;
  }
  /* Count black pixels in a row */
  for (x = 0; x < NORM_WIDTH; x++) {
    src = normed_digit + x;
    black = 0;
    for (y = 0; y < NORM_HEIGHT; y++) {
      if (*src) black++;
      src += NORM_WIDTH;
    }
    digit_vector[NORM_HEIGHT + x] = black;
  }
#ifdef VECTOR_DEBUG
  for (x = 0; x < NORM_WIDTH + NORM_HEIGHT; x++)
    cons_printf("%d, ", digit_vector[x]);
  cons_printf("\r\n");
#endif
}

/* Compute distance between digit dictionary and normalized bitmap */
short compute_distance(int number, int match)
{
  const short *dict;
  short *digit;
  int n, distance, diff;

  dict = digit_dictionary[number].vector;
  digit = digit_vector;
  distance = 0;
  for (n = 0; n < NORM_HEIGHT + NORM_WIDTH; n++) {
    diff = *digit++ - *dict++;
    distance += diff * diff;
  }
  diff = NORM_WIDTH * NORM_HEIGHT - match;
  distance += diff * diff;
  return distance;
}
#endif

void check_bars(IMAGEDATA *base, LABELPROP *pProp)
{
  IMAGEDATA *img;
  int w, h;
  int bbars, max_bars;
  uint8_t exp;

  max_bars = 0;
  for (w = 0; w < pProp->width; w++) {
    img = base + IMG_WIDTH * pProp->ypos + pProp->xpos + w;
    bbars = 0;
    exp = 0;
    for (h = 0; h < pProp->height; h++) {
      if (img->pix_label.pixel != exp) {
       if (exp == 0) bbars++;
       exp ^= 1;
      }
      img += IMG_WIDTH;
    }
    if (bbars > max_bars) max_bars = bbars;
  }
  pProp->ybars = max_bars;
  max_bars = 0;
  for (h = 0; h < pProp->height; h++) {
    img = base + IMG_WIDTH * (pProp->ypos + h) + pProp->xpos;
    bbars = 0;
    exp = 0;
    for (w = 0; w < pProp->width; w++) {
      if (img->pix_label.pixel != exp) {
       if (exp == 0) bbars++;
       exp ^= 1;
      }
      img ++;
    }
    if (bbars > max_bars) max_bars = bbars;
  }
  pProp->xbars = max_bars;
}

void find_best_match(LABELPROP *pProp)
{
  int max, best, mind;
#ifdef USE_VECTOR
  short val;
#endif
  int match, l, n;
  const uint8_t *p1, *p2;
  char digitstr[2];

  mind = 9999;
  max = 0;
  best = -1;
  for (n = 0; n < 10; n++) {
    p1 = &digit_images[n][0];
    p2 = normed_digit;
    match = 0;
    for (l = 0; l < NORM_WIDTH * NORM_HEIGHT; l++) {
      if (*p1 == *p2) match++;
      p1++;
      p2++;
    }
#ifdef USE_VECTOR
    val = compute_distance(n, match);
    cons_printf("%d: %d, %d\r\n", n, match, val);
    if (val < mind) {
      mind = val;
      max = match;
      best = n;
    }
#else
    if (match > max) {
      max = match;
      best = n;
    }
#endif
  }
  if (max < 80) {
    cons_printf("---- INVALID (best = %d, max = %d) ---\r\n", best, max);
    return;
  }
  cons_printf("----- %d (%d) -----\r\n", best, max);
  digitstr[0] = best + '0';
  digitstr[1] = 0;
  lcd_string(digitstr, FT_LARGE, pProp->xpos, 110, C_WHITE, C_BLACK);
}

void wait_key(OV_INFO *pinfo)
{
  if (!pinfo->direct_mode) {
    pinfo->navsw = 0;
    while (pinfo->navsw == 0)
      delay_ms(10);
  }
}

#define	FIX0299	77	/* 0.299 * 256 = 76.544 */
#define	FIX0587 150	/* 0.587 * 256 = 150.272 */
#define	FIX0114 29	/* 0.114 * 256 = 29.184 */

#define	HIST_AREA_X0 40
#define	HIST_AREA_Y0 50
#define	HIST_AREA_X1 120
#define	HIST_AREA_Y1 70

/*
 * Convert to gray scale picture
 */
void gray_scale(OV_INFO *pinfo, IMAGEDATA *img)
{
  uint8_t *pRaster;
  uint8_t red, green, blue;
  uint16_t yval, pixel;
  int x, y;

  for (y = 0; y < IMG_HEIGHT; y++) {   /* Check every pixels on the screen */
    for (x = 0; x < IMG_WIDTH; x++) {
      /* Convert RGB565 value into RGB888 format */
      red = img->rgb[0] & 0xf8;
      green = img->rgb[0] << 5 | img->rgb[1] >> 3; 
      blue = img->rgb[1] << 3;
      /*
       * Y = RED * 0.299 + GREEN * 0.587 + BLUE * 0.114
       */
      yval = red * FIX0299
            + green * FIX0587
            + blue * FIX0114;
      yval >>= 8;
      img->pix_label.label = yval;
      /* Collect histogram info, when we are scanning screen center area */
      if (y >= HIST_AREA_Y0 && y <= HIST_AREA_Y1 && x >= HIST_AREA_X0 && x < HIST_AREA_X1) {
        histogram[yval >> 3]++;
      }
      img++;
    }
    if (!pinfo->direct_mode) {	/* Display gray scale image */
      img -= IMG_WIDTH;
      pRaster = raster;
      for (x = 0; x < IMG_WIDTH; x++) {
        yval = img->pix_label.label;
        /* Convert into RGB565 format */
        pixel = ((yval & 0xf8) << 8) | ((yval & 0xfc) << 3)
            | ((yval & 0xf8) >> 3);
        *pRaster++ = pixel >> 8;
        *pRaster++ = pixel & 0xff;
        img++;
      }
      lcd_copy_line(raster, y);  /* Display this line */
    }
  }
}

static LABELPROP prop;

/*
* Discard object regions which doesn't look numbers.
*/
int discard_region(IMAGEDATA *img_buffer, int maxlabel)
{
  LABELPROP *pProp;
  int n, flag;
  int newmax;

  pProp = &prop;
  newmax = 0;
  for (n = 1; n < maxlabel; n++) {
    if (label_cleared[n/8] & (1<<(n&7))) {
      continue;
    }
    get_label_property(img_buffer, n, pProp);
    flag = 0;
    if (pProp->width > IMG_WIDTH/4 || pProp->width < 3)
      flag++;
    else if (pProp->height < NORM_HEIGHT)
      flag++;
    else if (pProp->ypos > IMG_HEIGHT/2)
      flag++;
    else if (pProp->ypos + pProp->height < IMG_HEIGHT/2)
      flag++;
    else if (pProp->width > pProp->height)
      flag++;
    if (flag) {
      label_clear(img_buffer, n, pProp);
      label_cleared[n/8] |= (1 << (n&7));
    } else {
      newmax = n;
    }
  }
  return newmax;
}

void number_decode(OV_INFO *pinfo)
{
  IMAGEDATA *img, *img_buffer;
  uint8_t color;
  int n, label_max;
  LABELPROP *pProp;
  uint8_t thres;
  int i, lowest, peak1, peak2, pdiff, diff;
  int hcount1, hcount2;

  memset(histogram, 0, sizeof(histogram));
  memset(histmark, 0, sizeof(histmark));
  memset(label_cleared, 0, sizeof(label_cleared));

  img_buffer = img = (IMAGEDATA *)ov_buffer;

  wait_key(pinfo);

  gray_scale(pinfo, img);

#ifdef H_DEBUG
  for (i = 0; i < N_HIST; i++) {
    hsum += histogram[i];
    cons_printf("%d ", histogram[i]);
  }
  cons_printf("\r\nSum = %d --> %d\r\n", hcount, hsum);
#endif

  wait_key(pinfo);

  /* Mark peak positions */
  pdiff = histogram[1] - histogram[0];
  if (pdiff > 0)
    histmark[0] = 'B';
  else if (pdiff < 0)
    histmark[0] = 'T';
  for (i = 1; i < N_HIST-1; i++) {
    diff = histogram[i+1] - histogram[i];
    if (diff < 0) {
      if (pdiff >= 0) {
        histmark[i] = 'T';
      }
    } else if (diff > 0) {
      if (pdiff <= 0) {
        histmark[i] = 'B';
      }
    }
    pdiff = diff;
  }
  if (pdiff > 0)
    histmark[i] = 'T';
  else if (pdiff < 0)
    histmark[i] = 'B';

  /* Find two peaks */
  peak1 = peak2 = 0;

  for (i = 0; i < N_HIST; i++) {
    if (histmark[i] == 'T') {
      if (histogram[i] > histogram[peak1]) {
        peak2 = peak1;
        peak1 = i;
      } else if (histogram[i] >= histogram[peak2]) {
        peak2 = i;
      }
    }
  }

  /*
   * Find bottom position located between two peaks
   *  and use that position as threshold value to distinguish black/white.
   */
  if (peak1 < peak2) {
    lowest = peak1;
    for (i = peak1; i < peak2; i++) {
      if (histmark[i] == 'B') {
        if (histogram[i] < histogram[lowest])
          lowest = i;
      }
    }
  } else {  /* peak1 >= peak2 */
    lowest = peak2;
    for (i = peak2; i < peak1; i++) {
      if (histmark[i] == 'B') {
        if (histogram[i] < histogram[lowest])
          lowest = i;
      }
    }
  }
  hcount1 = hcount2 = 0;
  for (i = 0; i < N_HIST; i++) {
    if (i < lowest) {
      hcount1 += histogram[i];
    } else {
      hcount2 += histogram[i];
    }
  }
#ifdef XDEBUG
  cons_printf("peak1: %d, peak2: %d, lowest: %d\r\n", peak1, peak2, lowest);
#endif
  thres = lowest * 8;

  /* Generate black/white image using determined threshold value */
  img_buffer = img = (IMAGEDATA *)ov_buffer;
  for (n = 0; n < IMG_WIDTH * IMG_HEIGHT; n++) {
    color = (img->pix_label.label >= thres)? WHITE_PIXEL : BLACK_PIXEL;
    if (hcount1 < hcount2)
      img->pix_label.pixel = color ^ 1;
    else
      img->pix_label.pixel = color & 1;
    img->pix_label.label = 0;
    img++;
  }

  if (!pinfo->direct_mode) {
    bmp_write(img_buffer);  /* Display B&W image */
    pinfo->navsw = 0;
    while (pinfo->navsw == 0)
      delay_ms(10);
  }

  label_max = labeling(img_buffer);

  label_max = discard_region(img_buffer, label_max);

  bmp_write(img_buffer);

  wait_key(pinfo);

  pProp = &prop;
  for (n = label_max; n > 0; n--) {
    if (label_cleared[n/8] & (1<<(n&7))) {
      continue;
    }
    get_label_property(img_buffer, n, pProp);
    if (pProp->area) {
      check_bars(img_buffer, pProp);
      cons_printf("%d: (%d, %d) --> %d X %d\r\n", n, pProp->xpos, pProp->ypos,
         pProp->width, pProp->height);
#ifdef BDEBUG
      cons_printf("xbar: %d, ybar: %d\r\n", pProp->xbars, pProp->ybars);
#endif
      if (pProp->xbars < 4 && pProp->ybars < 5) {
        resize_region(img_buffer, pProp);
#ifdef USE_VECTOR
        compute_vector();
#endif
        find_best_match(pProp);
      }
    }
  }

  pinfo->navsw = 0;
  while (pinfo->navsw == 0)
    delay_ms(10);
}
