/*
 * OmniVision OV7670 CMOS Camera driver for ATMEL SAM3S
 *
 * Written by Toshiharu Ohno (sirius506@gmail.com)
 */
#include <string.h>
#include <stdlib.h>
#include "SAMCamera.h"
#include "i2c.h"

#ifndef NEW_HEADER
/*
 * Old ATMEL SAM3S header file dones't have DMA related register definitions
 * for parallel capture function. Here, we add these registers.
 */
typedef struct {
  RwReg PIO_RPR;        /**< \brief (Pio Offset: 0x168) Receive Pointer Register */
  RwReg PIO_RCR;        /**< \brief (Pio Offset: 0x16C) Receive Counter Register */
  RwReg PIO_TPR;        /**< \brief (Pio Offset: 0x170) Transmit Pointer Register */
  RwReg PIO_TCR;        /**< \brief (Pio Offset: 0x174) Transmit Counter Register */
  RwReg PIO_RNPR;       /**< \brief (Pio Offset: 0x178) Receive Next Pointer Register */
  RwReg PIO_RNCR;       /**< \brief (Pio Offset: 0x17C) Receive Next Counter Register */
  RwReg PIO_TNPR;       /**< \brief (Pio Offset: 0x180) Transmit Next Pointer Register */
  RwReg PIO_TNCR;       /**< \brief (Pio Offset: 0x184) Transmit Next Counter Register */
  WoReg PIO_PTCR;       /**< \brief (Pio Offset: 0x188) Transfer Control Register */
  RoReg PIO_PTSR;       /**< \brief (Pio Offset: 0x18C) Transfer Status Register */
} PioPdc;

#define	PDC_PIOA	CAST(PioPdc, 0x400E0F68U)
#endif

/*
 * OV7670 is directory connected to SAM3S4B
 * using Parallel Capture as shown below.
 *
 *   OV7670                           SAM3S4B
 *
 *   SCL   ------------------------>  PA3   TWD0
 *   SDA   ------------------------>  PA4   TWCK0
 *   XCLK  <------------------------  PA6   PCK0
 *  VSYNC  ------------------------>  PA15  PIODCEN1
 *   HREF  ------------------------>  PA16  PIODCEN2
 *   PCLK  ------------------------>  PA23  PIODCCLK
 *    D0   ------------------------>  PA24  PIODC0
 *    D1   ------------------------>  PA25  PIODC1
 *    D2   ------------------------>  PA26  PIODC2
 *    D3   ------------------------>  PA27  PIODC3 
 *    D4   ------------------------>  PA28  PIODC4
 *    D5   ------------------------>  PA29  PIODC5
 *    D6   ------------------------>  PA30  PIODC6
 *    D7   ------------------------>  PA31  PIODC7
 */

#define	VSYNC_BIT PIO_PA15

uint8_t ov_buffer[OV_BUFSIZE];     /* Frame buffer memory */
OV_INFO ov_info;
ov_configs *pconfigs;

extern ov_configs ov7670_configs;
#ifdef USE_OV7725
extern ov_configs ov7725_configs;
#endif

/*
 * There are two reasons to generate PIOA interrupt.
 *
 * 1. Vertical Sync Interrupt (detect rising edge)
 * 2. Pallarel Capture ``End of DMA'' interrupt
 *
 */
void PIOA_IRQHandler()
{
  Pio *pPIOA;
  PioPdc *pPDC;
  uint32_t isr;
  OV_INFO *pinfo = &ov_info;

  pPIOA = PIOA;
  pPDC = PDC_PIOA;

  /*
   * Check pin change interrupt caused by VSYNC
   */
  isr = pPIOA->PIO_ISR & pPIOA->PIO_IMR;
  if (isr) {	/* Vsync rising edge */
    pinfo->vsync_count++;
    if (pinfo->status == ST_PREVIEW) {
      if (pinfo->dma_busy) {	/* DMA transfer is in progress. Ignore this Vsync */
        pinfo->vsync_ignore++;
      } else { /* No DMA activity, start it. */
        /* Prepare for Parallel capture DMA transfer */ 
#ifdef NEW_HEADER
#if OV_PIX_BYTES == 1
        pPIOA->PIO_PCMR = PIO_PCMR_HALFS | PIO_PCMR_DSIZE(2);	/* DSIZE = 2 (32bit) */
#else
        pPIOA->PIO_PCMR = PIO_PCMR_DSIZE(2);	/* DSIZE = 2 (32bit) */
#endif
        pPIOA->PIO_RPR = (RwReg) ov_buffer;
        /* Read one frame image at once */
        pPIOA->PIO_RCR = OV_BUFSIZE/4;
        pPIOA->PIO_PTCR = PIO_PTCR_RXTEN;
#else
#if OV_PIX_BYTES == 1
        pPIOA->PIO_PCMR = (1 << 10) | (2 << 4);	/* DSIZE = 2 (32bit) */
#else
        pPIOA->PIO_PCMR = (2 << 4);		/* DSIZE = 2 (32bit) */
#endif
        pPDC->PIO_RPR = (RwReg) ov_buffer;
        /* Read one frame image at once */
        pPDC->PIO_RCR = OV_BUFSIZE/4;
        pPDC->PIO_PTCR = PERIPH_PTCR_RXTEN;
#endif
        pPIOA->PIO_PCIER = PIO_PCIER_ENDRX;	/* Enable end of DMA interrupt */
        pPIOA->PIO_PCMR |= PIO_PCMR_PCEN;	/* Start Capture */
        pinfo->dma_busy++;
      }
    }
  }
  /*
   * End of DMA means we have received one frame
   */
  isr = pPIOA->PIO_PCISR & pPIOA->PIO_PCIMR;
  if (isr & PIO_PCISR_ENDRX) {
    pPIOA->PIO_PCIDR = PIO_PCIDR_ENDRX; /* Stop further interrupt request */
#ifdef NEW_HEADER
    pPIOA->PIO_PTCR = PIO_PTCR_RXTDIS;  /* Stop DMA activity */
#else
    pPDC->PIO_PTCR = PERIPH_PTCR_RXTDIS;  /* Stop DMA activity */
#endif
    pPIOA->PIO_PCMR = 0;
    if (pinfo->status == ST_PREVIEW) {
      pinfo->event |= CAMEV_DMA_DONE;
    }
  }
}

uint8_t ov_read_reg(uint32_t regno)
{
  return i2c_read_reg(OV_ADDR, regno);
}

void ov_write_reg(uint32_t regno, uint8_t bval)
{
  i2c_write_reg(OV_ADDR, regno, bval);
  delay_ms(OV_DELAY);	/* It looks we need some delay! */
}

/*
 * Configure SAM3S PIOA pins to meet with our hardware requirements.
 */
void configure_pio()
{
  Pio *pPIOA = PIOA;

  pPIOA->PIO_IFER = VSYNC_BIT;		/* Enable glitch filter on VSYNC */
  pPIOA->PIO_ESR = VSYNC_BIT;		/* Edge select */
  pPIOA->PIO_REHLSR = VSYNC_BIT;	/* Rising edge */
  pPIOA->PIO_AIMER = VSYNC_BIT;		/* Advanced interrupt mode */

  pPIOA->PIO_IDR = VSYNC_BIT;		/* Disable Vsync interrupt */
  pPIOA->PIO_PCIDR = PIO_PCIDR_ENDRX;   /* disable interrupt request */

  NVIC_EnableIRQ(PIOA_IRQn);		/* Enable PIOA interrupt */
}

#define	PID_OV7725 0x7721
#define	PID_OV7670 0x7673
#define	MID_OMNIVISION 0x7fa2

/*
 * See if we have expected Omnivision camera connected.
 * Currently, OV7670 is the only one we support.
 */
ov_configs *ov_detect()
{
  uint16_t mid, pidver;
  ov_configs *pconfigs;

  i2c_init();
  configure_pio();

  mid = (ov_read_reg(0x1c) << 8);
  delay_ms(OV_DELAY);
  mid |= ov_read_reg(0x1d);
  delay_ms(OV_DELAY);
  pidver = (ov_read_reg(0x0a) << 8);
  delay_ms(OV_DELAY);
  pidver |= ov_read_reg(0x0b);
  if (mid != MID_OMNIVISION) {	/* Not a OMNIVISION Sensor */
    return NULL;
  }
  switch (pidver)  {
  case PID_OV7670:	/* OV7670 detected */
    pconfigs = &ov7670_configs;
    break;
#ifdef USE_OV7725
  case PID_OV7725:
    pconfigs = &ov7725_configs;
    break;
#endif
  default:
    pconfigs = NULL;
    break;
  }
  return pconfigs;
}

/*
 * Initialize the Ominivision sensor with given configuration.
 */
void ov_init(const ov_config *cp)
{
  Pio *pPIOA = PIOA;
  PioPdc *pPDC;
  int n = 0;

  pPDC = PDC_PIOA;

  /* Disable Vsync interrupt and stops under going
   * DMA activity, if any.
   */
  pPIOA->PIO_IDR = VSYNC_BIT;
#ifdef NEW_HEADER
  pPIOA->PIO_RCR = 0;
  pPIOA->PIO_PTCR = PIO_PTCR_RXTDIS;  /* Stop DMA activity */
#else
  pPDC->PIO_RCR = 0;
  pPDC->PIO_PTCR = PERIPH_PTCR_RXTDIS;  /* Stop DMA activity */
#endif

  while (cp->regno != 0xff) {
    if (cp->regno == 0xfe) {
      delay_ms(cp->regval * 10);
    } else {
      ov_write_reg(cp->regno, cp->regval);
      n++;
    }
    cp++;
  }
}

void ov_start(OV_INFO *pinfo)
{
  Pio *pPIOA = PIOA;

  pinfo->vsync_ignore = pinfo->vsync_count = 0;
  pPIOA->PIO_IER = VSYNC_BIT;   /* Enable VSYNC interrupt */
  pinfo->status = ST_PREVIEW;
}

/* Stop image capture */
void ov_stop(OV_INFO *pinfo)
{
  Pio *pPIOA = PIOA;

  if (pinfo->status == ST_PREVIEW) {
     /* wait until undergoind DMA finished */
     while (!(pinfo->event & CAMEV_DMA_DONE))
       ;
     pinfo->dma_busy = 0;
  }
  pPIOA->PIO_IDR = VSYNC_BIT;	/* Disable Vsync interrupt */
  pinfo->status = ST_STOPPED;
}

void ov_restart(OV_INFO *pinfo)
{
  Pio *pPIOA = PIOA;

  pPIOA->PIO_IER = VSYNC_BIT;	/* Enable Vsync interrupt */
  pinfo->status = ST_PREVIEW;
  pinfo->need_clear = 1;
}

void ov_control(int swval)
{
  OV_INFO *pinfo = &ov_info;

  pinfo->event |= CAMEV_NAVSW;
  pinfo->navsw = swval;
}

void ov_set_resolution(RESOLUTION res)
{
  switch (res) {
  case RES_VGA:
    ov_init(pconfigs->vga_config);
    break;
  case RES_QVGA:
    ov_init(pconfigs->qvga_config);
    break;
  default: /* Treat as RES_QCIF */
    ov_init(pconfigs->qcif_config);
    break;
  }
}

void ov_set_exposure(OV_INFO *pinfo)
{
  ov_write_reg(0x13, 0xe2); /* Disable AGC and AEC */
  ov_write_reg(0x00, pinfo->gain);  /* Gain */
  ov_write_reg(0x10, pinfo->exposure);  /* Exposure */
}
