/*
 * SPI interface
 *
 * SPI interface is used to connect LCD at NPCS0.
 */
#include <string.h>
#include "SAMCamera.h"
#include "spi.h"
#include "lcd.h"

static volatile int sem_spi_rx, sem_spi_tx;

void SPI_IRQHandler()
{
  volatile Spi *pSPI;
  uint32_t sr;

  pSPI = SPI;

  sr = pSPI->SPI_SR & pSPI->SPI_IMR;
  if (sr & SPI_SR_TXBUFE) {
    pSPI->SPI_IDR = SPI_IDR_TXBUFE;
    pSPI->SPI_PTCR = SPI_PTCR_TXTDIS;
    sem_spi_tx = 1;
  }

  if (sr & SPI_SR_RXBUFF) {
    pSPI->SPI_IDR = SPI_IDR_RXBUFF;
    pSPI->SPI_PTCR = SPI_PTCR_RXTDIS | SPI_PTCR_TXTDIS;
    sem_spi_rx = 1;
  }
}

void spi_init(void)
{
  volatile Spi *pSPI  = SPI;
  volatile Pio *pPIOA = PIOA;

  /* Configure PIO */
  pPIOA->PIO_PDR = PIO_PA11 | PIO_PA12 | PIO_PA13 | PIO_PA14;
  pPIOA->PIO_ABCDSR[0] &= ~(PIO_PA11A_NPCS0 | PIO_PA12A_MISO | PIO_PA13A_MOSI | PIO_PA14A_SPCK);
  pPIOA->PIO_ABCDSR[1] &= ~(PIO_PA11A_NPCS0 | PIO_PA12A_MISO | PIO_PA13A_MOSI | PIO_PA14A_SPCK);

  /* PA10 as LCD_RS */
  pPIOA->PIO_OER = PIO_PA10;

  NVIC_EnableIRQ(SPI_IRQn);	/* Enable SPI interrupt */

  pSPI->SPI_CR = SPI_CR_SWRST | SPI_CR_SPIEN;	// Software reset, SPI Enable (0x81)
  pSPI->SPI_CR = SPI_CR_SPIEN;			// SPI Enable (0x01)

  /* SPI_CSR[0] for LCD */
  pSPI->SPI_CSR[0] =
	(0x0 << 24) |	// Delay between Consecutive Transfers (0 MCK periods)
	(0x0 << 16)  |	// Delay before SPCK (1 MCK period)
	(2 << 8)    |	// Serial Clock Baud Rate (MCK/2 = 64/2 = 32MHz)
	SPI_CSR_BITS_8_BIT |	// Bits per Transfer (8 bits)
	SPI_CSR_NCPHA    	// Clock Phase (data captured on raising edge)
        ;

  pSPI->SPI_MR =
	(0 << 24) |	// Dley between CS
	(0x0e << 16)  |	// Peripheral Chip Select (selects SPI_NPCS0)
	SPI_MR_MODFDIS |	// Mode Fault Detection (disabled)
	SPI_MR_MSTR;   		// Master Mode
}

void spi_send_only(uint8_t *bufp, int cnt)
{
  volatile Spi *pSPI = SPI;

  while ((pSPI->SPI_SR & SPI_SR_TXEMPTY) == 0);

  pSPI->SPI_MR =
	    (0 << 24) |	// Delay between CS
	    (0x0e << 16)  |	// Peripheral Chip Select (selects SPI_NPCS0)
	    SPI_MR_MODFDIS |	// Mode Fault Detection (disabled)
	    SPI_MR_MSTR;   	// Master Mode

  pSPI->SPI_TPR = (RwReg) bufp;
  pSPI->SPI_TCR = (RwReg) cnt;
  pSPI->SPI_IER = SPI_IER_TXBUFE;
  pSPI->SPI_PTCR = SPI_PTCR_TXTEN;
  while (sem_spi_tx == 0)
    ;
  sem_spi_tx = 0;
}

void spi_send(uint8_t dev, uint8_t *bufp, uint8_t *result, int cnt)
{
  volatile Spi *pSPI = SPI;

  while ((pSPI->SPI_SR & SPI_SR_TXEMPTY) == 0);

  pSPI->SPI_MR =
	    (0 << 24) |	// Delay between CS
	    (0x0e << 16)  |	// Peripheral Chip Select (selects SPI_NPCS0)
	    SPI_MR_MODFDIS |	// Mode Fault Detection (disabled)
	    SPI_MR_MSTR;   	// Master Mode

  pSPI->SPI_TPR = (RwReg) bufp;
  pSPI->SPI_TCR = (RwReg) cnt;
  pSPI->SPI_RPR = (RwReg) result;
  pSPI->SPI_RCR = (RwReg) cnt;
  pSPI->SPI_IER = SPI_IER_RXBUFF;
  pSPI->SPI_PTCR = SPI_PTCR_TXTEN | SPI_PTCR_RXTEN;

  /* Wait for DMA done interrupt */
  while (sem_spi_rx == 0)
    ;
  sem_spi_rx = 1;
}

void spi_recv(uint8_t *result, int cnt)
{
  volatile Spi *pSPI = SPI;

  while ((pSPI->SPI_SR & SPI_SR_TXEMPTY) == 0);

  // Select SPI mode SD card
  pSPI->SPI_MR =
	    (2 << 24) |	// Delay between CS
	    (0x07 << 16)   |	// Peripheral Chip Select (selects SPI_NPCS3)
	    SPI_MR_MODFDIS |	// Mode Fault Detection (disabled)
	    SPI_MR_MSTR;   	// Master Mode

  pSPI->SPI_TPR = (RwReg) result;
  pSPI->SPI_TCR = (RwReg) cnt;
  pSPI->SPI_RPR = (RwReg) result;
  pSPI->SPI_RCR = (RwReg) cnt;
  pSPI->SPI_IER = SPI_IER_RXBUFF;
  pSPI->SPI_PTCR = SPI_PTCR_TXTEN | SPI_PTCR_RXTEN;

  /* Wait for DMA done interrupt */
  while (sem_spi_rx == 0)
    ;
  sem_spi_rx = 1;
}
